# -*- coding: utf-8 -*-
"""
:Module:            freshpy.errors.handlers
:Synopsis:          Functions that handle various error situations within the namespace
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     28 Dec 2021
"""

import sys


def eprint(*args, **kwargs):
    """This function behaves the same as the ``print()`` function but is leveraged to print errors to ``sys.stderr``."""
    print(*args, file=sys.stderr, **kwargs)
    return



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.core
:Synopsis:          Defines the core freshpy object used to interface with the Freshservice API
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     04 Jan 2022
"""

from . import api, errors
from . import tickets as tickets_module
from .utils import log_utils, version

# Initialize logging
logger = log_utils.initialize_logging(__name__)


class FreshPy(object):
    """This is the class for the core object leveraged in this library."""
    # Define the function that initializes the object instance (i.e. instantiates the object)
    def __init__(self, domain=None, api_key=None):
        """This method instantiates the core Fresh object.

        .. versionadded:: 1.0.0
        """
        # Define the current version
        self.version = version.get_full_version()

        # Raise an exception if the domain and API key were not supplied
        if not domain or not api_key:
            raise errors.exceptions.MissingRequiredDataError('init')

        # Define the domain
        domain = f'https://{domain}' if domain and not domain.startswith('http') else domain
        domain = domain[:-1] if domain.endswith('/') else domain
        self.domain = domain

        # Define the base URL
        self.base_url = f'{domain}/api/v2/'

        # Define the API key
        self.api_key = api_key

        # Import inner object classes so their methods can be called from the primary object
        self.tickets = self._import_tickets_class()

    def _import_tickets_class(self):
        """This method allows the :py:class:`freshpy.core.FreshPy.Tickets` class to be utilized in the core object.

        .. versionadded:: 1.0.0
        """
        return FreshPy.Tickets(self)

    def get(self, uri, headers=None, return_json=True, verify_ssl=True):
        """This method performs a GET request against the Freshservice API with multiple retries on failure.

        .. versionchanged:: 1.1.0
           Added the ability to disable SSL verification on API calls.

        .. versionadded:: 1.0.0

        :param uri: The URI to query
        :type uri: string
        :param headers: The HTTP headers to utilize in the REST API call
        :type headers: dict, None
        :param return_json: Determines if JSON data should be returned
        :type return_json: bool
        :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
        :type verify_ssl: bool
        :returns: The JSON data from the response or the raw :py:mod:`requests` response.
        :raises: :py:exc:`freshpy.errors.exceptions.APIConnectionError`
        """
        return api.get_request_with_retries(self, uri, headers, return_json, verify_ssl=verify_ssl)

    class Tickets(object):
        """This class includes methods associated with Freshservice tickets."""
        def __init__(self, freshpy_object):
            """This method initializes the :py:class:`freshpy.core.freshpy.Tickets` inner class object.

            .. versionadded:: 1.0.0

            :param freshpy_object: The core :py:class:`freshpy.FreshPy` object
            :type freshpy_object: class[freshpy.FreshPy]
            """
            self.freshpy_object = freshpy_object

        def get_ticket(self, ticket_number, include=None, verify_ssl=True):
            """This method returns the data for a specific ticket.

            .. versionchanged:: 1.1.0
               Added the ability to disable SSL verification on API calls.

            .. versionadded:: 1.0.0

            :param ticket_number: The ticket number for which to return data
            :type ticket_number: str, int
            :param include: A string or iterable of `embedding <https://api.freshservice.com/#view_a_ticket>`_ options
            :type include: str, tuple, list, set, None
            :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
            :type verify_ssl: bool
            :returns: JSON data for the given ticket
            :raises: :py:exc:`freshpy.errors.exceptions.APIConnectionError`
            """
            return tickets_module.get_ticket(self.freshpy_object, ticket_number, include, verify_ssl=verify_ssl)

        def get_tickets(self, include=None, predefined_filter=None, filters=None, filter_logic='AND', requester_id=None,
                        requester_email=None, ticket_type=None, updated_since=None, ascending=None, descending=None,
                        per_page=None, page=None, verify_ssl=True):
            """This method returns a sequence of tickets with optional filters.

            .. versionchanged:: 1.1.0
               Added the ability to disable SSL verification on API calls.

            .. versionadded:: 1.0.0

            :param include: A string or iterable of `embedding <https://api.freshservice.com/#view_a_ticket>`_ options
            :type include: str, tuple, list, set, None
            :param predefined_filter: One of the predefined filters ('new_and_my_open', 'watching', 'spam', 'deleted')
            :type predefined_filter: str, None
            :param filters: Query filter(s) in the form of a structured query string or a dictionary of values
            :type filters: str, dict, None
            :param filter_logic: Defines the logic to use as necessary in a filter query string (default is ``AND``)
            :param requester_id: The numeric ID of a requester
            :param requester_id: The numeric ID of a requester
            :type requester_id: str, int, None
            :param requester_email: The email address of a requester
            :type requester_email: str, None
            :param ticket_type: The type of ticket (e.g. ``Incident``, ``Service Request``, etc.)
            :type ticket_type: str, None
            :param updated_since: A threshold date or timestamp (in UTC format) for when the ticket was last updated
            :type updated_since: str, None
            :param ascending: Determines if the tickets should be sorted in *ascending* order
            :type ascending: bool, None
            :param descending: Determines if the tickets should be sorted in *descending* order (default)
            :type descending: bool, None
            :param per_page: Displays a certain number of results per query
            :type per_page: str, int, None
            :param page: Returns a specific page number (used for paginated results)
            :type page: str, int, None
            :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
            :type verify_ssl: bool
            :returns: A list of JSON objects for tickets
            :raises: :py:exc:`freshpy.errors.exceptions.InvalidPredefinedFilterError`,
                     :py:exc:`freshpy.errors.exceptions.APIConnectionError`
            """
            return tickets_module.get_tickets(self.freshpy_object, include=include, predefined_filter=predefined_filter,
                                              filters=filters, filter_logic=filter_logic, requester_id=requester_id,
                                              per_page=per_page, page=page, requester_email=requester_email,
                                              ticket_type=ticket_type, updated_since=updated_since, ascending=ascending,
                                              descending=descending, verify_ssl=verify_ssl)

    def __del__(self):
        """This method fully destroys the instance.

        .. versionadded:: 1.0.0
        """
        self.close()

    def close(self):
        """This core method destroys the instance.

        .. versionadded:: 1.0.0
        """



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.errors.exceptions
:Synopsis:          Collection of exception classes relating to the freshpy library
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     28 Dec 2021
"""

#################
# Base Exception
#################


# Define base exception class
class FreshPyError(Exception):
    """This is the base class for FreshPy exceptions.

    .. versionadded:: 1.0.0
    """
    pass


############################
# Authentication Exceptions
############################


class MissingAuthDataError(FreshPyError):
    """This exception is used when authentication data is not supplied and therefore a connection cannot occur.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The authentication data was not provided and a connection cannot be established."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


#####################
# General Exceptions
#####################


class CurrentlyUnsupportedError(FreshPyError):
    """This exception is used when a feature or functionality being used is currently unsupported.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "This feature is currently unsupported at this time."
        if not (args or kwargs):
            args = (default_msg,)
        else:
            custom_msg = f"The '{args[0]}' {default_msg.split('This ')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class DataMismatchError(FreshPyError):
    """This exception is used when there is a mismatch between two data sources.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "A data mismatch was found with the data sources."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'data' in kwargs:
            multi_types = [list, tuple, set]
            if type(kwargs['data']) == str:
                custom_msg = f"{default_msg.split('data')[0]}'{kwargs['val']}'{default_msg.split('with the')[1]}"
                custom_msg = custom_msg.replace('sources', 'source')
                args = (custom_msg,)
            elif type(kwargs['data']) in multi_types and len(kwargs['data']) == 2:
                custom_section = f"'{kwargs['data'][0]}' and '{kwargs['data'][1]}'"
                custom_msg = f"{default_msg.split('data sources')[0]}{custom_section}{default_msg.split('with the')[1]}"
                args = (custom_msg,)
        super().__init__(*args)


class InvalidFieldError(FreshPyError):
    """This exception is used when an invalid field is provided.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The field that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('field ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class InvalidURLError(FreshPyError):
    """This exception is used when a provided URL is invalid.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The provided URL is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'url' in kwargs:
            custom_msg = f"{default_msg.split('is')[0]}'{kwargs['url']}'{default_msg.split('URL')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class MissingRequiredDataError(FreshPyError):
    """This exception is used when a function or method is missing one or more required arguments.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "Missing one or more required parameters"
        init_msg = "The object failed to initialize as it is missing one or more required arguments."
        param_msg = "The required parameter 'PARAMETER_NAME' is not defined"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'init' in args or 'initialize' in args:
            if 'object' in kwargs:
                custom_msg = f"{init_msg.split('object')[0]}'{kwargs['object']}'{init_msg.split('The')[1]}"
                args = (custom_msg,)
            else:
                args = (init_msg,)
        elif 'param' in kwargs:
            args = (param_msg.replace('PARAMETER_NAME', kwargs['param']),)
        else:
            args = (default_msg,)
        super().__init__(*args)


#########################
# Generic API Exceptions
#########################


class APIConnectionError(FreshPyError):
    """This exception is used when the API query could not be completed due to connection aborts and/or timeouts.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The API query could not be completed due to connection aborts and/or timeouts."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class APIRequestError(FreshPyError):
    """This exception is used for generic API request errors when there isn't a more specific exception.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The API request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class DELETERequestError(FreshPyError):
    """This exception is used for generic DELETE request errors when there isn't a more specific exception.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The DELETE request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)
        
        
class FeatureNotConfiguredError(FreshPyError):
    """This exception is used when an API request fails because a feature is not configured.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        exc_msg = "The feature is not configured."
        if 'identifier' in kwargs or 'feature' in kwargs:
            if 'identifier' in kwargs:
                exc_msg += f" Identifier: {kwargs['identifier']}"
            if 'feature' in kwargs:
                exc_msg = exc_msg.replace("feature", f"{kwargs['feature']} feature")
            args = (exc_msg,)
        elif not (args or kwargs):
            args = (exc_msg,)
        super().__init__(*args)


class GETRequestError(FreshPyError):
    """This exception is used for generic GET request errors when there is not a more specific exception.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The GET request did not return a successful response."
        custom_msg = "The GET request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)
        
        
class InvalidPayloadValueError(FreshPyError):
    """This exception is used when an invalid value is provided for a payload field.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "An invalid payload value was provided."
        custom_msg = "The invalid payload value 'X' was provided for the 'Y' field."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'value' in kwargs:
            if 'field' in kwargs:
                custom_msg = custom_msg.replace('X', kwargs['value']).replace('Y', kwargs['field'])
            else:
                custom_msg = f"{custom_msg.replace('X', kwargs['value']).split(' for the')[0]}."
            args = (custom_msg,)
        super().__init__(*args)


class InvalidRequestTypeError(FreshPyError):
    """This exception is used when an invalid API request type is provided.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied request type for the API is not recognized. (Examples of valid " + \
                      "request types include 'POST' and 'PUT')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class LookupMismatchError(FreshPyError):
    """This exception is used when an a lookup value does not match the supplied lookup type.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The supplied lookup type for the API does not match the value that was provided."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class NotFoundResponseError(FreshPyError):
    """This exception is used when an API query returns a 404 response and there isn't a more specific class.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The API query returned a 404 response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class POSTRequestError(FreshPyError):
    """This exception is used for generic POST request errors when there isn't a more specific exception.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The POST request did not return a successful response."
        custom_msg = "The POST request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class PUTRequestError(FreshPyError):
    """This exception is used for generic PUT request errors when there isn't a more specific exception.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "The PUT request did not return a successful response."
        custom_msg = "The PUT request failed with the following message:"
        if 'status_code' in kwargs or 'message' in kwargs:
            if 'status_code' in kwargs:
                status_code_msg = f"returned the {kwargs['status_code']} status code"
                custom_msg = custom_msg.replace('failed', status_code_msg)
            if 'message' in kwargs:
                custom_msg = f"{custom_msg} {kwargs['message']}"
            else:
                custom_msg = custom_msg.split(' with the following')[0] + "."
            args = (custom_msg,)
        elif not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


####################
# Ticket Exceptions
####################


class InvalidFilterLogicError(FreshPyError):
    """This exception is used when an invalid filter logic operator is supplied.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "An invalid filter logic operator was provided."
        custom_msg = "The filter logic operator 'X' is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'value' in kwargs:
            custom_msg = custom_msg.replace('X', kwargs['value'])
            args = (custom_msg,)
        super().__init__(*args)


class InvalidPredefinedFilterError(FreshPyError):
    """This exception is used when the API query could not be completed due to connection aborts and/or timeouts.

    .. versionadded:: 1.0.0
    """
    def __init__(self, *args, **kwargs):
        """This method defines the default or custom message for the exception."""
        default_msg = "An invalid predefined filter was provided."
        custom_msg = "The provided filter 'X' is not a valid predefined filter."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'value' in kwargs:
            custom_msg = custom_msg.replace('X', kwargs['value'])
            args = (custom_msg,)
        super().__init__(*args)



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.utils.version
:Synopsis:          This simple script contains the package version
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     04 Jan 2022
"""

from . import log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)

# Define special and global variables
__version__ = "1.1.0b1"


def get_full_version():
    """This function returns the current full version of the ``freshpy`` package.

    .. versionadded:: 1.0.0
    """
    return __version__


def get_major_minor_version():
    """This function returns the current major.minor (i.e. X.Y) version of the ``freshpy`` package."""
    return ".".join(__version__.split(".")[:2])



# -*- coding: utf-8 -*-
"""
:Package:           freshpy
:Synopsis:          This is the ``__init__`` module for the freshpy package
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     30 Dec 2021
"""

from .core import FreshPy
from .utils import version

__all__ = ['core', 'FreshPy', 'api', 'tickets']

# Define the package version by pulling from the freshpy.utils.version module
__version__ = version.get_full_version()




# -*- coding: utf-8 -*-
"""
:Package:        freshpy.utils
:Synopsis:       This is the ``__init__`` module for the freshpy.utils modules
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  27 Dec 2021
"""

__all__ = ['log_utils', 'version']



# -*- coding: utf-8 -*-
"""
:Package:        freshpy.errors
:Synopsis:       This module includes custom exceptions
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  28 Dec 2021
"""

__all__ = ['exceptions', 'handlers']

from . import exceptions, handlers



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.api
:Synopsis:          This module handles interactions with the Freshservice REST API
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     04 Jan 2022
"""

import json

import requests

from . import errors
from .utils import log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)


def define_headers():
    """This function defines the headers to use in API calls.

    .. versionadded:: 1.0.0
    """
    headers = {'Content-Type': 'application/json'}
    return headers


def define_auth(api_key):
    """This function defines the authentication dictionary to use in API calls.

    .. versionadded:: 1.0.0
    """
    credentials = (api_key, 'X')
    return credentials


def get_request_with_retries(fresh_object, uri, headers=None, return_json=True, verify_ssl=True):
    """This function performs a GET request and will retry several times if a failure occurs.

    .. versionchanged:: 1.1.0
       Added the ability to disable SSL verification on API calls.

    .. versionadded:: 1.0.0

    :param fresh_object: The instantiated :py:class:`freshpy.core.FreshPy` object.
    :param uri: The URI to query
    :type uri: string
    :param headers: The HTTP headers to utilize in the REST API call
    :type headers: dict, None
    :param return_json: Determines if JSON data should be returned
    :type return_json: bool
    :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
    :type verify_ssl: bool
    :returns: The JSON data from the response or the raw :py:mod:`requests` response.
    :raises: :py:exc:`freshpy.errors.exceptions.APIConnectionError`
    """
    # Define headers if not supplied
    headers = define_headers() if not headers else headers

    # Construct the credentials dictionary
    credentials = define_auth(fresh_object.api_key)

    # Construct the query URL
    query_url = fresh_object.base_url + uri

    # Perform the API call
    retries, response = 0, None
    while retries <= 5:
        try:
            response = requests.get(query_url, headers=headers, auth=credentials, verify=verify_ssl)
            break
        except Exception as exc_msg:
            _report_failed_attempt(exc_msg, 'get', retries)
            retries += 1
    if retries == 6:
        _raise_exception_for_repeated_timeouts()
        pass
    if return_json:
        response = response.json()
    return response


def _report_failed_attempt(_exc_msg, _request_type, _retries):
    """This function reports a failed API call that will be retried.

    .. versionadded:: 1.0.0

    :param _exc_msg: The exception that was raised within a try/except clause
    :param _request_type: The type of API request (e.g. ``post``, ``put`` or ``get``)
    :type _request_type: str
    :param _retries: The attempt number for the API request
    :type _retries: int
    :returns: None
    """
    _exc_name = type(_exc_msg).__name__
    if 'connect' not in _exc_name.lower():
        raise Exception(f"{_exc_name}: {_exc_msg}")
    _current_attempt = f"(Attempt {_retries} of 5)"
    _error_msg = f"The {_request_type.upper()} request has failed with the following exception: " + \
                 f"{_exc_name}: {_exc_msg} {_current_attempt}"
    errors.handlers.eprint(f"{_error_msg}\n{_exc_name}: {_exc_msg}\n")
    return


def _raise_exception_for_repeated_timeouts():
    """This function raises an exception when all API attempts (including) retries resulted in a timeout.

    .. versionadded:: 1.0.0

    :returns: None
    :raises: :py:exc:`freshpy.errors.exceptions.APIConnectionError`
    """
    _failure_msg = "The script was unable to complete successfully after five consecutive API timeouts. " + \
                   "Please run the script again or contact Freshservice Support for further assistance."
    raise errors.exceptions.APIConnectionError(_failure_msg)



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.tickets
:Synopsis:          Functions for interacting with Freshservice tickets
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     04 Jan 2022
"""

from . import api, errors
from .utils import core_utils, log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)

# Define constants
VALID_PREDEFINED_FILTERS = ['new_and_my_open', 'watching', 'spam', 'deleted']
SUPPORTED_FILTER_FIELDS = ['agent_id', 'group_id', 'priority', 'status', 'impact', 'urgency', 'tag', 'due_by',
                           'fr_due_by', 'created_at']
FILTER_LOGIC_OPERATORS = ['AND', 'OR']


def get_ticket(freshpy_object, ticket_number, include=None, verify_ssl=True):
    """This function returns the data for a specific ticket.

    .. versionchanged:: 1.1.0
       Added the ability to disable SSL verification on API calls.

    .. versionadded:: 1.0.0

    :param freshpy_object: The core :py:class:`freshpy.FreshPy` object
    :type freshpy_object: class[freshpy.FreshPy]
    :param ticket_number: The ticket number for which to return data
    :type ticket_number: str, int
    :param include: A string or iterable of `embedding <https://api.freshservice.com/#view_a_ticket>`_ options
    :type include: str, tuple, list, set, None
    :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
    :type verify_ssl: bool
    :returns: JSON data for the given ticket
    :raises: :py:exc:`freshpy.errors.exceptions.APIConnectionError`
    """
    uri = f'tickets/{ticket_number}'
    uri += _parse_constraints(_include=include)
    return api.get_request_with_retries(freshpy_object, uri, verify_ssl=verify_ssl)


def get_tickets(freshpy_object, include=None, predefined_filter=None, filters=None, filter_logic='AND',
                requester_id=None, requester_email=None, ticket_type=None, updated_since=None, ascending=None,
                descending=None, per_page=None, page=None, verify_ssl=True):
    """This function returns a sequence of tickets with optional filters.

    .. versionchanged:: 1.1.0
       Added the ability to disable SSL verification on API calls.

    .. versionadded:: 1.0.0

    :param freshpy_object: The core :py:class:`freshpy.FreshPy` object
    :type freshpy_object: class[freshpy.FreshPy]
    :param include: A string or iterable of `embedding <https://api.freshservice.com/#view_a_ticket>`_ options
    :type include: str, tuple, list, set, None
    :param predefined_filter: One of the predefined filters ('new_and_my_open', 'watching', 'spam', 'deleted')
    :type predefined_filter: str, None
    :param filters: Query filter(s) in the form of a structured query string or a dictionary of values
    :type filters: str, dict, None
    :param filter_logic: Defines the logic to use as necessary in a filter query string (default is ``AND``)
    :param requester_id: The numeric ID of a requester
    :type requester_id: str, int, None
    :param requester_email: The email address of a requester
    :type requester_email: str, None
    :param ticket_type: The type of ticket (e.g. ``Incident``, ``Service Request``, etc.)
    :type ticket_type: str, None
    :param updated_since: A date or timestamp (in UTC format) to be a threshold for when the ticket was last updated
    :type updated_since: str, None
    :param ascending: Determines if the tickets should be sorted in *ascending* order
    :type ascending: bool, None
    :param descending: Determines if the tickets should be sorted in *descending* order (default)
    :type descending: bool, None
    :param per_page: Displays a certain number of results per query
    :type per_page: str, int, None
    :param page: Returns a specific page number (used for paginated results)
    :type page: str, int, None
    :param verify_ssl: Determines if SSL verification should occur (``True`` by default)
    :type verify_ssl: bool
    :returns: A list of JSON objects for tickets
    :raises: :py:exc:`freshpy.errors.exceptions.InvalidPredefinedFilterError`,
             :py:exc:`freshpy.errors.exceptions.APIConnectionError`
    """
    uri = 'tickets'
    if filters:
        uri += _parse_filters(filters, filter_logic)
    else:
        uri += _parse_constraints(_include=include, _predefined_filter=predefined_filter, _requester_id=requester_id,
                                  _requester_email=requester_email, _ticket_type=ticket_type,
                                  _updated_since=updated_since, _ascending=ascending, _descending=descending,
                                  _per_page=per_page, _page=page)
    return api.get_request_with_retries(freshpy_object, uri, verify_ssl=verify_ssl)


def _parse_filters(_filters=None, _logic='AND'):
    _filters = {} if not _filters else _filters
    if _logic.upper() not in FILTER_LOGIC_OPERATORS:
        raise errors.exceptions.InvalidFilterLogicError(value=_logic)
    if isinstance(_filters, str):
        _filters = core_utils.url_encode(_filters)
        _uri_segment = f'/filter?query="{_filters}"'
    else:
        _uri_segment = f'/filter?query='
        _filter = ''
        for _idx, (_field, _value) in enumerate(_filters.items()):
            _filter += f'{_field}:{_value}'
            if _idx < (len(_filters) - 1):
                _filter += f' {_logic.upper()} '
        _filter = core_utils.url_encode(_filter)
        _uri_segment += f'"{_filter}"'
    return _uri_segment


def _parse_constraints(_include=None, _predefined_filter=None, _requester_id=None, _requester_email=None,
                       _ticket_type=None, _updated_since=None, _ascending=None, _descending=None, _per_page=None,
                       _page=None):
    """This function parses any constraints into a properly constructed query string.

    .. versionadded:: 1.0.0

    :param _include: A string or iterable of `embedding <https://api.freshservice.com/#view_a_ticket>`_ options
    :type _include: str, tuple, list, set, None
    :param _predefined_filter: One of the predefined filters ('new_and_my_open', 'watching', 'spam', 'deleted')
    :type _predefined_filter: str, None
    :param _requester_id: The numeric ID of a requester
    :type _requester_id: str, int, None
    :param _requester_email: The email address of a requester
    :type _requester_email: str, None
    :param _ticket_type: The type of ticket (e.g. ``Incident``, ``Service Request``, etc.)
    :type _ticket_type: str, None
    :param _updated_since: A date or timestamp (in UTC format) to be a threshold for when the ticket was last updated
    :type _updated_since: str, None
    :param _ascending: Determines if the tickets should be sorted in *ascending* order
    :type _ascending: bool, None
    :param _descending: Determines if the tickets should be sorted in *descending* order (default)
    :type _descending: bool, None
    :param _per_page: Displays a certain number of results per query
    :type _per_page: str, int, None
    :param _page: Returns a specific page number (used for paginated results)
    :type _page: str, int, None
    :returns: The fully constructed query string
    :raises: :py:exc:`freshpy.errors.exceptions.InvalidPredefinedFilterError`
    """
    _constraints = ''
    if _include:
        # TODO: Perform a check to verify support for the provided include value(s)
        if not isinstance(_include, str):
            _include = ','.join(_include)
        _constraints = core_utils.construct_query_string(_constraints, f'include={_include}')
    if _predefined_filter:
        if _predefined_filter not in VALID_PREDEFINED_FILTERS:
            if isinstance(_predefined_filter, str):
                raise errors.exceptions.InvalidPredefinedFilterError(value=_predefined_filter)
            raise errors.exceptions.InvalidPredefinedFilterError()
        _constraints = core_utils.construct_query_string(_constraints, f'filter={_predefined_filter}')
    if _requester_id:
        _constraints = core_utils.construct_query_string(_constraints, f'requester_id={_requester_id}')
    if _requester_email:
        _requester_email = core_utils.url_encode(_requester_email)
        _constraints = core_utils.construct_query_string(_constraints, f'requester_email={_requester_email}')
    if _ticket_type:
        _ticket_type = _ticket_type.replace(' ', '+')
        _constraints = core_utils.construct_query_string(_constraints, f'type={_ticket_type}')
    if _updated_since:
        _constraints = core_utils.construct_query_string(_constraints, f'updated_since={_updated_since}')
    if _ascending:
        _constraints = core_utils.construct_query_string(_constraints, 'order_type=asc')
    if _descending:
        _constraints = core_utils.construct_query_string(_constraints, 'order_type=desc')
    if _per_page:
        _constraints = core_utils.construct_query_string(_constraints, f'per_page={_per_page}')
    if _page:
        _constraints = core_utils.construct_query_string(_constraints, f'page={_page}')
    return _constraints




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:Synopsis:          This script is the primary configuration file for the freshpy project
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     04 Jan 2022
"""

import setuptools
import codecs
import os.path


def read(rel_path):
    """This function reads the ``version.py`` script in order to retrieve the version.

    .. versionadded:: 1.0.0
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """This function retrieves the current version of the package without needing to import the
       :py:mod:`freshpy.utils.version` module in order to avoid dependency issues.

    .. versionadded:: 1.0.0
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delimiter = '"' if '"' in line else "'"
            return line.split(delimiter)[1]
    raise RuntimeError("Unable to find the version string")


with open('README.md', 'r') as fh:
    long_description = fh.read()

version = get_version('src/freshpy/utils/version.py')

setuptools.setup(
    name='freshpy',
    version=version,
    author='Jeff Shurtliff',
    author_email='jeff.shurtliff@rsa.com',
    description='A Python toolset for utilizing the Freshservice API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jeffshurtliff/freshpy",
    project_urls={
        'Change Log': 'https://freshpy.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/jeffshurtliff/freshpy/issues',
    },
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Communications",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Office/Business",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires='>=3.6',
    install_requires=[
        "urllib3>=1.26.7",
        "requests>=2.26.0",
        "setuptools~=52.0.0"
    ],
    extras_require={
        'sphinx': [
            'Sphinx>=3.4.0',
            'sphinxcontrib-applehelp>=1.0.2',
            'sphinxcontrib-devhelp>=1.0.2',
            'sphinxcontrib-htmlhelp>=1.0.3',
            'sphinxcontrib-jsmath>=1.0.1',
            'sphinxcontrib-qthelp>=1.0.3',
            'sphinxcontrib-serializinghtml>=1.1.4'
        ],
    }
)



# -*- coding: utf-8 -*-
"""
:Module:            freshpy.utils.core_utils
:Synopsis:          Collection of supporting utilities and functions to complement the primary modules
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     30 Dec 2021
"""

import urllib.parse


def construct_query_string(existing_query=None, appendage=None):
    """This function assists in constructing query strings for URIs to ensure they follow the appropriate format.

    .. versionadded:: 1.0.0

    :param existing_query: The existing query string (if any)
    :type existing_query: str, None
    :param appendage: The new addition to the query string to be appended (if any)
    :type appendage: str, None
    :returns: The constructed query string
    """
    combined_query = '' if existing_query is None else existing_query
    if appendage:
        if existing_query == '?':
            combined_query = f'?{appendage}'
        elif existing_query:
            combined_query = existing_query + f'&{appendage}'
        elif appendage.startswith('?'):
            combined_query = appendage
        else:
            combined_query = f'?{appendage}'
    return combined_query


def url_encode(raw_string):
    """This function encodes a string for use in URLs.

    .. versionadded:: 1.0.0

    :param raw_string: The raw string to be encoded
    :type raw_string: str
    :returns: The encoded string
    """
    return urllib.parse.quote_plus(raw_string)


def url_decode(encoded_string):
    """This function decodes a url-encoded string.

    .. versionadded:: 1.0.0

    :param encoded_string: The url-encoded string
    :type encoded_string: str
    :returns: The unencoded string
    """
    return urllib.parse.unquote_plus(encoded_string)